/**
 * Driver program for the Histogram Equalization assignment.  This program is
 * responsible for loading and display images and for reacting to button
 * presses.
 * 
 * @author Keith Schwarz (htiek@cs.stanford.edu)
 */

import java.awt.event.*;
import java.io.*;
import java.util.Arrays;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import acm.program.*;
import acm.util.*;
import acm.graphics.*;

public class HistogramEqualization extends GraphicsProgram implements ComponentListener {
	/* The image being displayed. */
	private GImage displayImage;
	
	/* The master image. */
	private GImage masterImage;
	
	/* The button for equalizing / restoring. */
	private JButton toggleButton;
	
	/* Valid file extensions for images that we can read. */
	private static final String[] LOAD_IMAGE_EXTENSIONS = new String[] {
		".png", ".bmp", ".wbmp", ".jpg", ".gif", ".jpeg"
	};

	/**
	 * Sets up the window.
	 */
	public void init() {	
		/* Add buttons for loading and enhancing images. */
		add(new JButton("Choose Image"), SOUTH);
		add(toggleButton = new JButton("Equalize"), SOUTH);
		
		setIsEqualize(true);
		
		/* Register event listeners. */
		addComponentListener(this);
		addActionListeners();
	}
	
	/**
	 * Reacts to button presses.
	 */
	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().equals("Choose Image")) {
			chooseImage();
		} else if (e.getActionCommand().equals("Equalize")) {
			equalizeImage();
		} else if (e.getActionCommand().equals("Original")) {
			restoreImage();
		}
	}
	
	/**
	 * Stretches and redraws the chosen image.
	 */
	private void redrawAll() {
		if (displayImage != null) {
			displayImage.setBounds(0, 0, getWidth(), getHeight());
		}
	}
	
	/**
	 * Runs the logic to equalize the image, then displays the result.
	 */
	private void computeEqualizedImage() {
		assert displayImage != null;
		
		/* First, get the underlying image intensities. */
		int[][] pixels =      HistogramEqualizationImageTransforms.toGrayscale(displayImage).getPixelArray();
		int[][] intensities = HistogramEqualizationImageTransforms.imageToLuminances(pixels);
		
		/* Pass this off to the logic. */
		int[][] result = HistogramEqualizationLogic.equalize(intensities);
		
		/* Validate basic properties of the result. */
		if (confirmResult(result, pixels)) {
			/* Convert this back to pixels. */
			for (int row = 0; row < result.length; row++) {
				for (int col = 0; col < result[row].length; col++) {
					int intensity = result[row][col];
					result[row][col] = GImage.createRGBPixel(intensity, intensity, intensity);
				}
			}
			
			setImage(new GImage(result));
		}
	}
	
	/**
	 * Verifies that the given result image is a valid image, reporting an error if it is not.
	 * 
	 * @param result The candidate result.
	 * @param pixels The original image.
	 */
	private boolean confirmResult(int[][] result, int[][] pixels) {
		/* Make sure the result is non-null. */
		if (result == null) {
			getDialog().showErrorMessage("Resulting image was null.");
			return false;
		}
		
		/* Check it has the right length. */
		if (result.length != pixels.length) {
			getDialog().showErrorMessage("Resulting image has the wrong number of rows.");
			return false;
		}
		
		/* Check each row individually. */
		for (int i = 0; i < result.length; i++) {
			/* Confirm this line is valid. */
			if (result[i] == null) {
				getDialog().showErrorMessage("Resulting image contains a null row.");
				return false;
			}
			
			if (result[i].length != pixels[i].length) {
				getDialog().showErrorMessage("Resulting image has the wrong number of columns.");
				return false;
			}
			
			/* Check all pixels to see that they're in range. */
			for (int j = 0; j < result[i].length; j++) {
				if (result[i][j] < 0 || result[i][j] >= 256) {
					getDialog().showErrorMessage("Luminance out of range: " + result[i][j]);
					return false;
				}
			}
		}
		
		return true;
	}
	
	/**
	 * Sets whether the toggle button equalizes or restores.
	 * 
	 * @param isEqualize Whether the button equalizes.
	 */
	private void setIsEqualize(boolean isEqualize) {
		toggleButton.setText(isEqualize? "Equalize" : "Original");
	}
	
	/**
	 * Sets which image is currently being displayed.
	 * 
	 * @param image The image to display, or null if no image should be displayed.
	 */
	private void setImage(GImage image) {
		if (displayImage != null) {
			remove(displayImage);
		}
		
		displayImage = image;
		
		if (image != null) {
			displayImage.setBounds(0, 0, 0, 0);
			add(displayImage);
		}
		
		redrawAll();
	}
	
	/**
	 * Given a filename, returns the extension of that filename, or "" if it does not
	 * have one.
	 * 
	 * @param filename The filename in question.
	 * @return Its extension, or "" if it has none.
	 */
	private String extensionOf(File filename) {
		/* Find the suffix; fail if none is found. */
		int lastDot = filename.getName().lastIndexOf('.');
		if (lastDot == -1) return "";
		
		/* Determine the extension. */
		return filename.getName().substring(lastDot);
	}
	
	/**
	 * Obtains a file chooser that can be used to select image files.
	 * @return
	 */
	private JFileChooser getFileChooser() {
		JFileChooser fc = new JFileChooser();
		fc.setFileFilter(new FileFilter() {
			@Override
			public boolean accept(File filename) {
				return filename.isDirectory() || Arrays.asList(LOAD_IMAGE_EXTENSIONS).contains(extensionOf(filename));
			}
			
			@Override
			public String getDescription() {
				return "Image files";
			}
		});
		
		/* Set the working directory to the current images directory. */
		fc.setCurrentDirectory(new File("histogram/"));
		
		return fc;
	}
	
	/**
	 * Prompts the user to choose an image file.
	 */
	private void chooseImage() {
		JFileChooser fc = getFileChooser();
		if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			try {
				/* Try loading the image, which might fail if the file was invalid. */
				masterImage = new GImage(fc.getSelectedFile().getAbsolutePath());
				setImage(masterImage);
				setIsEqualize(true);
			} catch (ErrorException err) {
				getDialog().showErrorMessage("Error loading file: " + err.getMessage());
			}
		}
	}
	
	/**
	 * Equalizes the image and displays the result.
	 */
	private void equalizeImage() {
		if (displayImage == null) {
			getDialog().showErrorMessage("No image selected.");
		} else {
			computeEqualizedImage();
			setIsEqualize(false);
		}
	}
	
	/**
	 * Restores the original image.
	 */
	private void restoreImage() {
		setImage(masterImage);
		setIsEqualize(true);
	}
	
	/**
	 * Resizes the displayed image when the window is resized
	 */
	public void componentResized(ComponentEvent arg0) {
		redrawAll();
	}
	
	/* Unused ComponentListener events. */
	public void componentHidden(ComponentEvent arg0) {}
	public void componentMoved(ComponentEvent arg0) {}
	public void componentShown(ComponentEvent arg0) {}
}
